# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`
__all__ = ['gen_fake_data', 'gen_init_val_files', 'prep_fit']