#TODO: Make this a yml with objectification in __init__
elliptical = {'params': ['ret0h', 'dreth',
                         'ret045', 'dret45',
                         'ret0r', 'dretr',
                         't_ret', 't_pol',
                         'I_sys', 'Q_in', 'U_in', 'V_in'],
              'global': ['Q_in', 'U_in', 'V_in']}
